/*
  This code was written by Raymond Chi-Wing Wong.
  Please read the following.

    Raymond Chi-Wing Wong, Yufei Tao, Ada Wai-Chee Fu and Xiaokui Xiao. 
    "On Efficient Spatial Matching", 
    the 33rd International Conference on Very Large Data Bases (VLDB'07), 
    Vienna, Austria on 23-28 Sept, 2007 

*/
#include "closestPairMatrix.h"

float **closest_findDistanceMatrix(twoData *aTwoData, int noOfDim)
{
	float **returnValue;

	int i, j;

	pointList *O;
	pointList *P;

	int indexO, indexP;

	int noOfPointO, noOfPointP;

	O = aTwoData->O;
	P = aTwoData->P;

	noOfPointO = O->noOfPoint;
	noOfPointP = P->noOfPoint;

	returnValue = (float **) malloc(sizeof(float *)*noOfPointO);
	for (i = 0; i < noOfPointO; i++)
	{
		returnValue[i] = (float *) malloc(sizeof(float)*noOfPointP);
		memset(returnValue[i], 0, sizeof(float)*noOfPointP);

		indexO = i;

		for (j = 0; j < noOfPointP; j++)
		{
			indexP = j;
			returnValue[i][j] = (float) stable_distance(indexO, indexP, O, P, noOfDim);
		}
	}

	return returnValue;
}

void closest_setInfinityForO(float **matrix, int indexO, int noOfPointP)
{
	int i;

	for (i = 0; i < noOfPointP; i++)
	{
		matrix[indexO][i] = (float) INFINITY;
	}
}

void closest_setInfinityForP(float **matrix, int indexP, int noOfPointO)
{
	int i;

	for (i = 0; i < noOfPointO; i++)
	{
		matrix[i][indexP] = (float) INFINITY;
	}
}

float closest_findMinForRow(float *aRow, int noOfPointP, int *returnIndexP)
{
	float returnValue;
	float minValue;
	int index;
	int i;

	float aValue;

	// assume noOfPointP >= 1
	minValue = aRow[0];
	index = 0;
	for (i = 1; i < noOfPointP; i++)
	{
		aValue = aRow[i];
		if (aValue < minValue)
		{
			minValue = aValue;
			index = i;
		}
	}

	returnValue = minValue;
	(*returnIndexP) = index;

	return returnValue;
}

void closest_findMinDistPair(float **matrix, int noOfPointO, int noOfPointP, int *indexO, int *indexP)
{
	int i;

	float *minDistForRow;
	int *indexMinDistForRow;
	int indexMinDistForAll;

	float *aRow;

	float minDistForAll;

	minDistForRow = (float *) malloc(sizeof(float)*noOfPointO);
	indexMinDistForRow = (int *) malloc(sizeof(int)*noOfPointO);
	memset(minDistForRow, 0, sizeof(float)*noOfPointO);
	memset(indexMinDistForRow, 0, sizeof(int)*noOfPointO);

	// find minDist
	minDistForAll = (float) INFINITY;
	indexMinDistForAll = -1;
	for (i = 0; i < noOfPointO; i++)
	{
		aRow = matrix[i];
		minDistForRow[i] = closest_findMinForRow(aRow, noOfPointP, &(indexMinDistForRow[i]));

		if (minDistForRow[i] < minDistForAll)
		{
			minDistForAll = minDistForRow[i];
			indexMinDistForAll = i;
		}
	}

	(*indexO) = indexMinDistForAll;
	(*indexP) = indexMinDistForRow[indexMinDistForAll];

	free(minDistForRow);
	free(indexMinDistForRow);
}

void closest_mainIteration(float **matrix, int noOfPointO, int noOfPointP, twoData *aTwoData, assignment *aAssignment)
{
	int indexO, indexP;

	int oid, pid;
	point *aPointO;
	point *aPointP;

	int matchWeight;

	triple *aTriple;

	closest_findMinDistPair(matrix, noOfPointO, noOfPointP, &indexO, &indexP);

	oid = indexO;
	pid = indexP + noOfPointO;

	aPointO = twoData_getOPoint(aTwoData, oid);
	aPointP = twoData_getPPoint(aTwoData, pid);

	if (aPointO->weight < aPointP->weight)
	{
		matchWeight = aPointO->weight;
	}
	else
	{
		matchWeight = aPointP->weight;
	}

	aPointO->weight -= matchWeight;
	aPointP->weight -= matchWeight;

	if (aPointO->weight == 0)
	{
		closest_setInfinityForO(matrix, indexO, noOfPointP);
	}

	if (aPointP->weight == 0)
	{
		closest_setInfinityForP(matrix, indexP, noOfPointO);
	}

	// assignment
	aTriple = triple_newValue(oid, pid, matchWeight);
	assignment_insert(aAssignment, &aTriple);
}

int closest_isEmpty(pointList *aPointList)
{
	int returnValue;

	int noOfPoint;
	point **pointList;
	point *aPoint;

	int i;

	noOfPoint = aPointList->noOfPoint;
	pointList = aPointList->list;

	returnValue = 1;
	for (i = 0; i < noOfPoint; i++)
	{
		aPoint = pointList[i];

		if (aPoint->weight != 0)
		{
			returnValue = 0;
			break;
		}
	}

	return returnValue;
}

assignment *closest_main(twoData *aTwoData, int noOfDim, float **matrix)
{
	assignment *returnValue;

	int noOfPointO;
	int noOfPointP;

	pointList *O;
	pointList *P;

	int isEmptyO, isEmptyP;

	twoData *cloneTwoData;

	cloneTwoData = twoData_clone(aTwoData, noOfDim);

	O = cloneTwoData->O;
	P = cloneTwoData->P;

	noOfPointO = O->noOfPoint;
	noOfPointP = P->noOfPoint;

	isEmptyO = closest_isEmpty(O);
	isEmptyP = closest_isEmpty(P);

	returnValue = assignment_new();

	while ((isEmptyP == 0) && (isEmptyO == 0))
	{
		// iteration
		closest_mainIteration(matrix, noOfPointO, noOfPointP, cloneTwoData, returnValue);

		// next iteratio
		isEmptyO = closest_isEmpty(O);
		isEmptyP = closest_isEmpty(P);
	}


	return returnValue;
}

